#ifndef C_P1CAMERA_SUBSCRIPTIONS_H
#define C_P1CAMERA_SUBSCRIPTIONS_H

#include "C_P1CameraCommonStructs.h"
#include <array>
#include <iosfwd>
#include <memory>
#include <vector>

namespace P1::CameraSdk {

/**
 * @brief Interface common to all subscriptions
 *
 * A subscription allows the camera to send you notifications in response to
 * events in the camera. Suscribing just means that you enable this event
 * delivery.
 *
 * To receive the notifications from the subscribed events, you *must* still
 * enable notifications for the actual events, related to the subscription.
 *
 * Further, when activating a subscription (by subscribing), you might change
 * state in the camera or enable certain hardware.
 *
 * @since Version 3.0
 */
struct ISubscription {
  /** @brief Check if the camera supports this subscription  */
  virtual bool IsSupported() const  = 0;
  /** @brief Check if we are currently subscribed  */
  virtual bool IsSubscribed() const = 0;
  /** @brief Begin subscribing and receive notifications for the related events*/
  virtual void Subscribe()          = 0;
  /**
   * @brief Cancel the subscription, no new notifications will be posted
   *
   * Note that there might still be a backlog of pending notifications from
   * previous posted events. Unsubscribing does *not* clear that queue.
   */
  virtual void Unsubscribe()        = 0;
};
std::ostream& operator<<(std::ostream& os, ISubscription const& subscription);

/**
 * @brief Subscribe / unsubscribe to preview transfers.
 *
 * Subscribing to this tells the camera that we want to receive previews of
 * newly captured images. The previews arrive via the notification
 * @ref EventType::CameraPreviewReady.
 *
 * @sa ISubscriptionManager
 * @since Version 3.0
 */
class IPreviewSubscription : public virtual ISubscription {

public:
  using SupportedFormats = std::vector<ImageFormat>;

  /**
   * @brief Subscribe to previews in the desired format.
   *
   * See @ref ImageFormatsSupported for supported formats
   */
  virtual void             Subscribe(ImageFormat format) = 0;

  /**
   * @brief Subscribe to previews using the format defined by @ref Format
   *
   * See @ref ImageFormatsSupported for supported formats for the camera
   */
  virtual void             Subscribe() = 0;
  /** @brief Formats supported for previews by the camera */
  virtual SupportedFormats ImageFormatsSupported() const = 0;
  /** @brief Get current format in use */
  virtual ImageFormat      Format() const                = 0;
  /** @brief Change the preview format */
  virtual void             SetFormat(ImageFormat format) = 0;
};
std::ostream& operator<<(std::ostream& os, IPreviewSubscription const& previewSubscription);

/**
 * @brief Subscribe / unsubscribe to a LiveView stream of images.
 *
 * Subscribing to this tells the camera that we would like a video stream of
 * images from the camera. Also using this class interface, LiveView
 * subscriptions can be configured for format and size.
 *
 * Subscribing to this subscription will activate LiveView the camera, and
 * stream image frames to your host, as long as the subscription is active.
 *
 * @sa ISubscriptionManager
 * @since Version 3.0
 */
class ILiveViewSubscription : public virtual ISubscription {
public:

  using SupportedFormats = std::vector<ImageFormat>;

  /** @brief Get supported formats that can be set with SetFormat()  */
  EXPORT_API virtual SupportedFormats ImageFormatsSupported() const = 0;

  /** @brief Subscribe to live view  */
  EXPORT_API virtual void Subscribe(ImageFormat format) = 0;

  /**
   * @brief Subscribe to previews using the format defined by @ref Format
   *
   * See @ref ImageFormatsSupported for supported formats for the camera
   */
  virtual void Subscribe() = 0;

  /** @brief Get the current bitmap format setting of LiveView */
  EXPORT_API virtual ImageFormat Format() const = 0;

  /** @brief Get the current crop applied to the LiveView */
  EXPORT_API virtual ValueRect Crop() const = 0;

  /** @brief Get the current desired max output dimension of the LiveView */
  EXPORT_API virtual uint32_t PreferredMaxDimension() const = 0;

  /**
   * @brief Helper method to reset a current crop, back to "no crop"
   *
   * If you "zoomed" in the LiveView by setting the crop, use this method to get the original
   * width and height of the full image.
   */
  EXPORT_API virtual ValueArea FullImageSize() const = 0;

  /**
   * @brief An abstract value for exposure compensation
   *
   * This values maps to the exposure compensation, but does not use the same EV values.
   * Instead it uses an opaque float point value.
   *
   * @return float The current lightness
   */
  EXPORT_API virtual float Lightness() const = 0;

  // Setters

  /** @brief Set the LiveView image's bitmap format.
   * Some formats might not be supported for older camera firmware versions.
   * To check that the value has been set correctly, poll with ImageFormatsSupported(). */
  EXPORT_API virtual void SetFormat(ImageFormat format) = 0;

  // /** @brief Set a new crop on the LiveView */
  EXPORT_API virtual void SetCrop(ValueRect const& cropRect) = 0;

  // /** @brief Reset the crop (to no crop), showing the entire image in the LIveView */
  EXPORT_API virtual void ResetCrop() = 0;

  // /** @brief Set the desired maximum dimension (width) of the LiveView's output images */
  EXPORT_API virtual void SetPreferredMaxDimension(uint32_t const maxDimensionValue) = 0;

  /**
   * @brief Set the *Custom C1* white balance, by the sampling a  position in full image coordinates
   *
   * Cameras like IQ4 defines a set of white balance settings. The defaults are *Daylight*, *Tungsten*,
   * etc. Besides the default settings, there are used definable ones. One of these are *Custom C1*.
   *
   * This method changes the cameras white balance property to *Custom C1* setting. And the
   * white balance vector for *Custom C1* is changed to reflect the sample position provided by this
   * method. **This means the cameras white balance setting is changed!**
   *
   * Use the *white balance* property, to change white balance back to *Auto* or *Daylight*, etc.
   */
  EXPORT_API virtual void SetWhiteBalancePoint(ValuePoint const& whitePoint) = 0;

  /** @brief Set a *exposure compensation*-like  value applied to the live view */
  EXPORT_API virtual void SetLightness(float const lightness) = 0;

  EXPORT_API virtual std::string ToString() const = 0;
};
std::ostream& operator<<(std::ostream& os, ILiveViewSubscription const& liveviewSubscription);

/**
 * @brief Subscribe / unsubscribe to IIQ image file transfers.
 *
 * Subscribing to this tells the camera that the host computer should ready
 * receive captures when the camera is triggered. The camera may choose to
 * ignore this (and not transmit any captures), as the storage configuration on
 * the camera controls the behaviour. (*Host Storage* must be enabled.)
 *
 * This subscription is also used by the *image browsing* feature, to send IIQ
 * files to the host. You *must* enable this subscription *before* calling
 * @ref IStoredImageHandle::RequestImageAsync
 *
 * @since Version 3.0
 */
class IFullImageSubscription : public virtual ISubscription {};
std::ostream& operator<<(std::ostream& os, IFullImageSubscription const& liveviewSubscription);

/**
 * @brief Access subscriptions available in the camera.
 *
 * Each @ref Camera object comes with a pre-initialized set of
 * @ref ISubscription objects, that allows control of different events posted
 * by the camera.
 *
 * Some subscriptions might not be available on all cameras. Unavailable
 * subscriptions will still be initialized, but @ref ISubscription::IsSupported
 * will return `false`.
 *
 * @since Version 3.0
 */
class ISubscriptionManager {
public:
  /**
   * @brief Subscription on receiving image previews when triggering captures
   */
  virtual std::shared_ptr<IPreviewSubscription>   Previews()   = 0;
  /**
   * @brief Subscription enabling and receiving LiveView image frames
   */
  virtual std::shared_ptr<ILiveViewSubscription>  LiveView()  = 0;
  /**
   * @brief Subscription on receiving full IIQ image files when triggering
   */
  virtual std::shared_ptr<IFullImageSubscription> FullImages() = 0;

};

} // namespace P1::CameraSdk

#endif // C_P1CAMERA_SUBSCRIPTIONS_H
